from starlight_vision import Unet3D, ElucidatedStarlight, StarlightTrainer

class Starlight:
    def __init__(self, 
                 dim=64, 
                 dim_mults=(1, 2, 4, 8), 
                 image_sizes=(16, 32),
                 random_crop_sizes=(None, 16),
                 temporal_downsample_factor=(2, 1),
                 num_sample_steps=10,
                 cond_drop_prob=0.1,
                 sigma_min=0.002,
                 sigma_max=(80, 160),
                 sigma_data=0.5,
                 rho=7,
                 P_mean=-1.2,
                 P_std=1.2,
                 S_churn=80,
                 S_tmin=0.05,
                 S_tmax=50,
                 S_noise=1.003):

        # Initialize the Unet models
        self.unet1 = Unet3D(dim=dim, dim_mults=dim_mults).cuda()
        self.unet2 = Unet3D(dim=dim, dim_mults=dim_mults).cuda()

        # Initialize the Starlight model
        self.starlight = ElucidatedStarlight(
            unets=(self.unet1, self.unet2),
            image_sizes=image_sizes,
            random_crop_sizes=random_crop_sizes,
            temporal_downsample_factor=temporal_downsample_factor,
            num_sample_steps=num_sample_steps,
            cond_drop_prob=cond_drop_prob,
            sigma_min=sigma_min,
            sigma_max=sigma_max,
            sigma_data=sigma_data,
            rho=rho,
            P_mean=P_mean,
            P_std=P_std,
            S_churn=S_churn,
            S_tmin=S_tmin,
            S_tmax=S_tmax,
            S_noise=S_noise,
        ).cuda()

        # Initialize the trainer
        self.trainer = StarlightTrainer(self.starlight)

    def train(self, videos, texts, unet_number=1, ignore_time=False):
        self.trainer(videos, texts=texts, unet_number=unet_number, ignore_time=ignore_time)
        self.trainer.update(unet_number=unet_number)

    def sample(self, texts, video_frames=20):
        return self.trainer.sample(texts=texts, video_frames=video_frames)




import torch 
import torch.nn as nn
from  torchvision.transforms import Compose, Resize, Normalize, ToTensor
from torch.utils.data import DataLoader
from transformers import DiffusionModel, ClipModel, DiffusionConfig, DPTImageProcessor, DPTForDepthEstimation
from torchvision.transforms import GaussianBlur
import torch.nn.functional as F
import math

#spatial transformer

class SpatialTransformer(nn.Module):
    def __init__(self, in_channels, num_heads):
        super(SpatialTransformer, self).__init__()
        self.in_channels = in_channels
        self.num_heads = num_heads

        self.key_proj = nn.Linear(in_channels, in_channels)
        self.value_proj = nn.Linear(in_channels, in_channels)
        self.query_proj = nn.Linear(in_channels, in_channels)
        self.softmax = nn.Softmax(dim=-1)
        self.output_proj = nn.Linear(in_channels, in_channels)

    def forward(self, x, content_embeddings):
        #compute keys and values from content embedding
        keys =  self.key_proj(content_embeddings)
        values = self.value_proj(content_embeddings)

        #compute queries from input
        queries = self.query_proj(x)


        #conpute attention scores
        attention_scores = torch.matmul(queries, keys.tranpose(-1. -2)) / (self.in_channels ** 0.5)
        attention_scores = self.softmax(attention_scores)


        #compute attented scores
        attented_values = torch.matmul(attention_scores, values)

        #add residual connection and apply output projection
        out = x + attented_values
        out = self.output_proj(out)

        return out
    

class Midas(nn.Module):
    def __init__(self):
        super(Midas, self).__init__()
        self.processor = DPTImageProcessor.from_pretrained("Intel/dpt-large")
        self.model = DPTForDepthEstimation.from_pretrained("Intel/dpt-large")
    
    def forward(self, x):
        #prepare images for the model
        inputs = self.processor(images=x, return_tensors="pt")

        with torch.no_grad():
            outputs = self.model(**inputs)
            predicted_depth = outputs.predicted_depth

        #interplate to original size
        prediction = torch.nn.function.interpolate(
            predicted_depth.unsqueeze(1),
            size=x.shape[-2:],
            mode="bicubic",
            align_corners=False,
        )

        return prediction.squeeze()
    


#sinusodial embeddings
class SinusoidalEmbedding(nn.Module):
    def __init__(self, num_channels):
        super(SinusoidalEmbedding, self).__init__()
        self.num_channels = num_channels

    def forward(self, ts, T):
        ts = torch.Tensor([ts])
        T = torch.Tensor([T])
        ts_embed = torch.zeros(self.num_channels // 2)
        div_term = torch.exp(torch.arange(0, self.num_channels // 2, 2) * -(math.log(10000.0) / self.num_channels))
        ts_embed[0::2] = torch.sin(ts * div_term)
        ts_embed[1::2] = torch.cos(ts * div_term)
        return ts_embed


#unet

class ResidualBlock(nn.Module):
    def __init__(self, in_channels, out_channels, temporal=False):
        super(ResidualBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.temporal = temporal
        if temporal:
            self.temporal_conv = nn.Conv1d(out_channels, out_channels, kernel_size=3, padding=1)
            self.bn_temporal = nn.BatchNorm1d(out_channels)

    def forward(self, x):
        residual = x
        out = self.relu(self.bn1(self.conv1(x)))
        if self.temporal:
            out = self.bn_temporal(self.temporal_conv(out))
        out = self.bn2(self.conv2(out))
        out += residual
        out = self.relu(out)
        return out
    

class TransformerBlock(nn.Module):
    def __init__(self, in_channels, temporal=False):
        super(TransformerBlock, self).__init__()
        self.spatial_transformer = nn.TransformerEncoderLayr(d_model=in_channels, nhead=8)
        self.temporal = temporal
        if temporal:
            self.temporal_transformer = nn.TransformerEncoderLayer(d_model=in_channels, nhead=8)

    def forward(self, x):
        out = self.spatial_transformer(x)
        if self.temporal:
            out = self.temporal_transformer(out)
        return out
    
class Unet(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(Unet, self).__init__()
        
        # Encoder
        self.encoder = nn.Sequential(
            ResidualBlock(in_channels, 64, temporal=False),
            nn.MaxPool2d(2),
            ResidualBlock(64, 128, temporal=False),
            nn.MaxPool2d(2),
            ResidualBlock(128, 256, temporal=False),
            nn.MaxPool2d(2),
            ResidualBlock(256, 512, temporal=False),
        )
        
        # Center part of the UNet
        self.middle = nn.Sequential(
            TransformerBlock(512, temporal=False),
        )

        # Decoder
        self.decoder = nn.Sequential(
            ResidualBlock(512, 256, temporal=False),
            nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True),
            ResidualBlock(256, 128, temporal=False),
            nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True),
            ResidualBlock(128, 64, temporal=False),
            nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True),
            nn.Conv2d(64, out_channels, kernel_size=1),
        )

    def forward(self, x):
        # Pass input through the encoder
        z = self.encoder(x)

        # Pass the latent representations through the middle UNet layers
        z = self.middle(z)
        
        # Pass output to decoder
        out = self.decoder(z)

        return out



#extended unet
class ExtendedUNet(Unet):
    def __init__(self, in_channels, out_channels, blur_kernel_size=3, blur_sigma=1.0):
        super(ExtendedUNet, self).__init__(in_channels, out_channels)
        self.midas_dpt_large = Midas() # Load the MiDaS DPT-Large model
        self.clip_model = ClipModel.from_pretrained("openai/clip-vit-base-patch32")
        self.blur = GaussianBlur(blur_kernel_size, sigma=blur_sigma)

    def process_structure(self, x, ts, Ts):
        depth_maps = self.midas_dpt_large(x)
        for _ in range(ts):
            depth_maps = self.blur(depth_maps)
            depth_maps = F.interpolate(depth_maps, scale_factor=0.5)
        depth_maps = F.interpolate(depth_maps, size=x.shape[-2:])
        z_structure = self.encoder(depth_maps)
        return z_structure

    def process_content(self, x):
        content_repr = self.clip_model.encode_image(x)
        return content_repr

    def sample(self, x, t, c, ts, guidance_scale=1.0, temporal_scale=1.0):
        zt = self.encoder(x)
        z_structure = self.process_structure(x, ts, Ts)
        zt = torch.cat([zt, z_structure], dim=1)
        content_repr = self.process_content(c)

        # Apply the spatial transformer and cross-attention conditioning
        out = self.layers(zt, content_repr)

        # Compute the adjusted predictions
        unconditional_pred = self.decoder(out)
        conditional_pred = self.decoder(out)
        adjusted_pred = unconditional_pred + guidance_scale * (conditional_pred - unconditional_pred)

        # Control temporal consistency
        image_model_pred = ... # Compute the prediction of the image model applied to each frame individually
        adjusted_pred = image_model_pred + temporal_scale * (adjusted_pred - image_model_pred)

        return adjusted_pred



class Starlight(nn.Module):
    def __init__(self, Ts):
        super(Starlight, self).__init__()

        self.unet_lowres = ExtendedUNet()
        self.unet_midres = ExtendedUNet()
        self.unet_highres = ExtendedUNet()
        self.sinusoidal_embedding = SinusoidalEmbedding(4)
        self.Ts = Ts
        self.lstm = nn.LSTM(input_size=512, hidden_size=512, num_layers=1)


    def forward(self, x, s, c, ts):
        #apply multi scale architecture
        x_lowres = F.interpolate(x, scale_factor=0.25)
        x_midres = F.interpolate(x, scale_factor=0.5)
        x_highres = x

        means_lowres = self.process_scale(x_lowres, s, c, ts)
        means_midres = self.process_scale(x_midres, s, c, ts)
        means_highres = self.process_scale(x_highres, s, c, ts)


        return means_lowres, means_midres, means_highres
    
    def process_structure(self, depth_maps, ts):
        for _ in range(ts):
            depth_maps = self.unet.blur(depth_maps)
            depth_maps = F.interpolate(depth_maps, scale_factor=0.5)
        depth_maps = F.interpolate(depth_maps, size=depth_maps.shape[-2:])
        z_structure = self.unet.encoder(depth_maps)
        return z_structure
    
    def process_scale(self, x, s, c, ts):
        # Compute content representation
        content_embedding = self.unet.process_content(x)

        # Compute structure representation
        depth_maps = self.unet.midas_dpt_large(x)
        structure_embedding = self.process_structure(depth_maps, ts)

        # Condition on structure
        zt = torch.cat([x, structure_embedding], dim=1)

        # Add sinusoidal embedding of ts
        ts_embed = self.sinusoidal_embedding(ts, self.Ts).unsqueeze(0).unsqueeze(-1).unsqueeze(-1)
        ts_embed = ts_embed.repeat(x.size(0), 1, x.size(2), x.size(3))
        zt = torch.cat([zt, ts_embed], dim=1)

        # Apply cross-attention
        zt = self.unet.spatial_transformer_blocks(zt, content_embedding, c)

        # Process LSTM temporal context
        lstm_input = zt.view(zt.size(0), 1, -1)
        _, (hidden_state, _) = self.lstm(lstm_input)
        hidden_state = hidden_state.view(hidden_state.size(1), -1, 1, 1)

        # Concatenate hidden_state with zt
        zt = torch.cat([zt, hidden_state], dim=1)

        # Pass through UNet
        means = self.unet(zt)

        return means
    
model = Starlight()

def train_starlight(model, train_loader, optimizer, device, perceptual_loss_fn, style_loss_fn, alpha, beta):
    model.train()

    for inputs, content, structure, ts in train_loader:
        inputs = inputs.to(device)
        content = content.to(device)
        structure = structure.to(device)
        ts = ts.to(device)

        optimizer.zero_grad()

        means_lowres, means_midres, means_highres = model(inputs, structure, content, ts)

        loss_lowres = F.mse_loss(inputs, means_lowres)
        loss_midres = F.mse_loss(inputs, means_midres)
        loss_highres = F.mse_loss(inputs, means_highres)

        # Perceptual and style losses
        perceptual_loss = perceptual_loss_fn(inputs, means_highres)
        style_loss = style_loss_fn(inputs, means_highres)

        loss = loss_lowres + loss_midres + loss_highres + alpha * perceptual_loss + beta * style_loss
        loss.backward()

        optimizer.step()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starlight_vision', 'starlight_vision.core']

package_data = \
{'': ['*']}

install_requires = \
['accelerate',
 'beartype',
 'click',
 'datasets',
 'einops>=0.6.1',
 'ema-pytorch>=0.0.3',
 'fsspec',
 'kornia',
 'lion-pytorch',
 'numpy',
 'packaging',
 'pillow',
 'pydantic>=2',
 'pytorch-lightning',
 'pytorch-warmup',
 'sentencepiece',
 'torch>=1.6',
 'torchvision',
 'tqdm',
 'transformers',
 'triton']

setup_kwargs = {
    'name': 'starlight-vision',
    'version': '0.1.1',
    'description': 'Starlight - unprecedented photorealism × deep level of language understanding',
    'long_description': "[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n\n# 🌌 Starlight Vision 🚀\n\n![Starlight](starlight.png)\n🪐 Starlight Vision is a powerful multi-modal AI model designed to generate high-quality novel videos using text, images, or video clips as input. By leveraging state-of-the-art deep learning techniques, it can synthesize realistic and visually impressive video content that can be used in a variety of applications, such as movie production, advertising, virtual reality, and more. 🎥\n\n## 🌟 Features\n\n- 📝 Generate videos from text descriptions\n- 🌃 Convert images into video sequences\n- 📼 Extend existing video clips with novel content\n- 🔮 High-quality output with customizable resolution\n- 🧠 Easy to use API for quick integration\n\n## 📦 Installation\n\nTo install Starlight Vision, simply use pip:\n\n```bash\npip install starlight-vision\n```\n\n## 🎬 Quick Start\n\nAfter we train you can install Starlight Vision and can start generating videos using the following code:\n\n```python\nimport torch\nfrom starlight_vision import Unet3D, ElucidatedStarlight, StarlightTrainer\n\nunet1 = Unet3D(dim = 64, dim_mults = (1, 2, 4, 8)).cuda()\n\nunet2 = Unet3D(dim = 64, dim_mults = (1, 2, 4, 8)).cuda()\n\n# elucidated starlight, which contains the unets above (base unet and super resoluting ones)\n\nstarlight = ElucidatedStarlight(\n    unets = (unet1, unet2),\n    image_sizes = (16, 32),\n    random_crop_sizes = (None, 16),\n    temporal_downsample_factor = (2, 1),        # in this example, the first unet would receive the video temporally downsampled by 2x\n    num_sample_steps = 10,\n    cond_drop_prob = 0.1,\n    sigma_min = 0.002,                          # min noise level\n    sigma_max = (80, 160),                      # max noise level, double the max noise level for upsampler\n    sigma_data = 0.5,                           # standard deviation of data distribution\n    rho = 7,                                    # controls the sampling schedule\n    P_mean = -1.2,                              # mean of log-normal distribution from which noise is drawn for training\n    P_std = 1.2,                                # standard deviation of log-normal distribution from which noise is drawn for training\n    S_churn = 80,                               # parameters for stochastic sampling - depends on dataset, Table 5 in apper\n    S_tmin = 0.05,\n    S_tmax = 50,\n    S_noise = 1.003,\n).cuda()\n\ntexts = [\n    'a whale breaching from afar',\n    'young girl blowing out candles on her birthday cake',\n    'fireworks with blue and green sparkles',\n    'dust motes swirling in the morning sunshine on the windowsill'\n]\n\nvideos = torch.randn(4, 3, 10, 32, 32).cuda() # (batch, channels, time / video frames, height, width)\n\n# feed images into starlight, training each unet in the cascade\n# for this example, only training unet 1\n\ntrainer = StarlightTrainer(starlight)\n\n# you can also ignore time when training on video initially, shown to improve results in video-ddpm paper. eventually will make the 3d unet trainable with either images or video. research shows it is essential (with current data regimes) to train first on text-to-image. probably won't be true in another decade. all big data becomes small data\n\ntrainer(videos, texts = texts, unet_number = 1, ignore_time = False)\ntrainer.update(unet_number = 1)\n\nvideos = trainer.sample(texts = texts, video_frames = 20) # extrapolating to 20 frames from training on 10 frames\n\nvideos.shape # (4, 3, 20, 32, 32)\n\n```\n\n## 🤝 Contributing\n\nWe welcome contributions from the community! If you'd like to contribute, please follow these steps:\n\n1. 🍴 Fork the repository on GitHub\n2. 🌱 Create a new branch for your feature or bugfix\n3. 📝 Commit your changes and push the branch to your fork\n4. 🚀 Create a pull request and describe your changes\n\n## 📄 License\n\nStarlight Vision is released under the APACHE License. See the [LICENSE](LICENSE) file for more details.\n\n## 🗺️ Roadmap\n\nThe following roadmap outlines our plans for future development and enhancements to Starlight Vision. We aim to achieve these milestones through a combination of research, development, and collaboration with the community.\n\n### 🚀 Short-term Goals\n\n- [ ] Improve text-to-video synthesis by incorporating advanced natural language understanding techniques\n- [ ] Train on LAION-5B and video datasets\n- [ ] Enhance the quality of generated videos through the implementation of state-of-the-art generative models\n- [ ] Optimize the model for real-time video generation on various devices, including mobile phones and edge devices\n- [ ] Develop a user-friendly web application that allows users to generate videos using Starlight Vision without any programming knowledge\n- [ ] Create comprehensive documentation and tutorials to help users get started with Starlight Vision\n\n### 🌌 Medium-term Goals\n\n- [ ] Integrate advanced style transfer techniques to allow users to customize the visual style of generated videos\n- [ ] Develop a plugin for popular video editing software (e.g., Adobe Premiere, Final Cut Pro) that enables users to utilize Starlight Vision within their existing workflows\n- [ ] Enhance the model's ability to generate videos with multiple scenes and complex narratives\n- [ ] Improve the model's understanding of object interactions and physics to generate more realistic videos\n- [ ] Expand the supported input formats to include audio, 3D models, and other media types\n\n### 🌠 Long-term Goals\n\n- [ ] Enable users to control the generated video with more granular parameters, such as lighting, camera angles, and object placement\n- [ ] Incorporate AI-driven video editing capabilities that automatically adjust the pacing, color grading, and transitions based on user preferences\n- [ ] Develop an API for real-time video generation that can be integrated into virtual reality, augmented reality, and gaming applications\n- [ ] Investigate methods for training Starlight Vision on custom datasets to generate domain-specific videos\n- [ ] Foster a community of researchers, developers, and artists to collaborate on the continued development and exploration of Starlight Vision's capabilities\n\n# Join Agora\nAgora is advancing Humanity with State of The Art AI Models like Starlight, join us and write your mark on the history books for eternity!\n\nhttps://discord.gg/sbYvXgqc\n\n\n\n## 🙌 Acknowledgments\n\nThis project is inspired by state-of-the-art research in video synthesis, such as the Structure and Content-Guided Video Synthesis with Diffusion Models paper, and leverages the power of deep learning frameworks like PyTorch.\n\nWe would like to thank the researchers, developers, and contributors who have made this project possible. 💫",
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/StarlightVision',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
